"""
Various tests.
"""

from env.common.run import run
from env.common.mktree import mktree


def test(tmp_path):
    mktree(path=tmp_path, dir_number=3, file_number=1, depth=3)

    if 1:  # Test 1: A test for a rule sequence via negation.
        (tmp_path / '.fls').write_text('**')
        (tmp_path / 'dirA/.fls').write_text('!**/*.log')
        (tmp_path / 'dirA/dirA/.fls').write_text('!.fls')
        (tmp_path / 'dirA/dirA/dirA/.fls').write_text('*.log')
        (tmp_path / 'dirA/dirB/.fls').write_text('!.fls')
        (tmp_path / 'dirB/.fls').write_text('!*')
        (tmp_path / 'dirB/dirB/.fls').write_text('!.fls')
        (tmp_path / 'dirB/dirB/dirB/.fls').write_text('!*.txt')
        (tmp_path / 'dirC/.fls').write_text('*.*')
        (tmp_path / 'dirC/dirC/.fls').write_text('!*.*')
        (tmp_path / 'dirC/dirC/dirC/.fls').write_text('*.*')

        test_cases = {
            'dirA': True,
            'dirA/file0.log': True,
            'dirA/file0.txt': True,
            'dirA/dirA/.fls': False,
            'dirA/dirA/file0.log': False,
            'dirA/dirA/file0.txt': True,
            'dirA/dirA/dirA/file0.log': True,
            'dirA/dirB/file0.log': False,
            'dirA/dirB/file0.txt': True,
            'dirB': True,
            'dirB/dirA/file0.txt': True,
            'dirB/dirB/dirB/file0.txt': False,
            'dirB/dirB/file0.txt': True,
            'dirB/dirB/.fls': False,
            'dirB/file0.txt': False,
            'dirC/dirC/file0.txt': False,
            'dirC/dirC/dirC/file0.txt': True,
            'dirC/.fls': True,
            'dirC/file0.txt': True,
            'file0.log': True,
            'file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='4')



"""
Module:        FLS
Description:   A module for matching file system paths against patterns.
Author:        Andrii Burkatskyi aka andr11b
Year:          2024
Version:       0.0.1.241224
License:       MIT License
Email:         4ndr116@gmail.com
Link:          https://github.com/codyverse/fls
"""

import os
import re
import random
import string
from collections import defaultdict


class FLS:
    """
    A class for managing file match patterns using custom `.fls` protocol files.

    This class scans directories starting from a root path, reads `.fls` files to load match rules, and
    determines whether files or directories should be matched based on those rules.

    Attributes:
        _root (str): The root directory where the scanning starts.
        _protocol (str): The filename of the protocol file to be processed (default: `.fls`).
        _rules (dict): A dictionary mapping directories to their associated rules.
        _matched (dict): A dictionary of file paths and whether they are matched.
    """

    class FLSRule:
        """
        Represents a single rule in the match protocol.

        Attributes:
            pattern (str): The raw pattern string from the match file.
            context (str): The directory context for the rule.
            regex (str): The compiled regular expression representing the rule.
            is_negation (bool): Whether the rule negates matching files.
            is_strictly_dir (bool): Whether the rule applies only to directories.
        """

        def __init__(self, pattern, context='', parent=None):
            if parent is None or not isinstance(parent, FLS):
                raise Exception("Cannot instantiate FLSRule directly.")

            self._rule = self._compose(pattern, context)

        def __str__(self):
            return self._rule

        def __repr__(self):
            return f'FLSRule({self._rule})'

        def _compose(self, pattern, context):
            placeholder_map = {}

            def _hidescape(match):
                placeholder = self._random()
                placeholder_map[placeholder] = match.group(0)

                return placeholder

            context = '' if context in {'.', '/'} else re.escape(context.rstrip('/')) + r'/'

            pattern = pattern.strip()

            repattern = re.sub(r'\\(.)', _hidescape, pattern)

            # Ignore empty lines and comments
            if repattern.strip() == '' or repattern[0] == '#':
                repattern = ''

            # Handle negation patterns (start with '!')
            if repattern.startswith('!'):
                is_negation = True
                repattern = repattern[1:]
            else:
                is_negation = False

            # Normalize slashes, e.g., `//` -> `/`, and remove leading slashes.
            repattern = (re.sub(r'/+', '/', repattern)).lstrip('/')
            #  Normalize asterisks, i.e., `***` -> `**`.
            repattern = re.sub(r'\*{3,}', '**', repattern)

            # Trailing slash (/) means this is explicitly a directory pattern
            is_strictly_dir = repattern.endswith('/')
            repattern = repattern.rstrip('/')

            if not is_strictly_dir and repattern == '**':
                repattern = r'.+$'
            elif not is_strictly_dir and repattern == '*':
                repattern = r'[^/]+/?$'
            else:

                def _rebrackets(match):
                    content = match.group(0)

                    if content.startswith('[!'): content = content.replace('[!', '[^')

                    placeholder = self._random()
                    placeholder_map[placeholder] = content

                    return f"{placeholder}"

                repattern = re.sub(r'\[.*?\]', _rebrackets, repattern)

                is_ending = False

                if not is_strictly_dir:
                    instance = r'/\*\*$'  # a trailing `/**`
                    if re.search(instance, repattern):
                        is_ending = True

                        placeholder = self._random()
                        placeholder_map[placeholder] = r'/(.+)$'
                        repattern = re.sub(instance, placeholder, repattern)

                    instance = r'\*\*$'  # a trailing `**`
                    if re.search(instance, repattern):
                        is_ending = True

                        placeholder = self._random()
                        placeholder_map[placeholder] = r'.*'
                        repattern = re.sub(instance, placeholder, repattern)

                    instance = r'/\*$'  # a trailing `/*`
                    if re.search(instance, repattern):
                        is_ending = True

                        placeholder = self._random()
                        placeholder_map[placeholder] = r'/[^/]+/?$'
                        repattern = re.sub(instance, placeholder, repattern)

                instance = r'\*\*'  # all instances of the `**`
                if re.search(instance, repattern):
                    placeholder = self._random()
                    placeholder_map[placeholder] = r'(.*)?'
                    repattern = re.sub(instance, placeholder, repattern)

                instance = r'\*'  # all instances of the `*`
                if re.search(instance, repattern):
                    placeholder = self._random()
                    placeholder_map[placeholder] = r'[^/]*'
                    repattern = re.sub(instance, placeholder, repattern)

                instance = r'\?'  # all instances of the the `?`
                if re.search(instance, repattern):
                    placeholder = self._random()
                    placeholder_map[placeholder] = r'[^/]{1}'
                    repattern = re.sub(instance, placeholder, repattern)

                repattern = re.escape(repattern)

                for placeholder, value in placeholder_map.items():
                    repattern = re.sub(re.escape(placeholder), value, repattern)

                if not is_ending:
                    if is_strictly_dir:
                        repattern += r'\/$'  # matches directory paths that explicitly end with a slash (/)
                    else:
                        repattern += r'(/?$)'  # matches paths that may or may not end with a slash (/)
            repattern = '^' + context + repattern

            return {
                'pattern': pattern,
                'context': context,
                'regex': repattern,
                'is_negation': is_negation,
                'is_strictly_dir': is_strictly_dir,
            }

        def _random(self, k=8):
            return f"_{''.join(random.choices(string.ascii_letters + string.digits, k=k))}_"

        def match(self, relpath):
            return bool(re.compile(self.regex).match(relpath))

        @property
        def rule(self):
            return self._rule

        @property
        def pattern(self):
            return self._rule['pattern']

        @property
        def regex(self):
            return self._rule['regex']

        @property
        def context(self):
            return self._rule['context']

        @property
        def is_strictly_dir(self):
            return self._rule['is_strictly_dir']

        @property
        def is_negation(self):
            return self._rule['is_negation']

    def __init__(self, root, protocol='.fls'):
        """
        Initialize the FLS instance.


        This method initializes the FLS class by setting the root directory from where the scanning starts and
        specifying the match protocol file.

        Args:
            root (str): The root directory where the scanning should begin.
                        This should be an absolute or relative path, and it must point to a valid directory.
                        If the path is relative, it will be resolved relative to the current working directory.
            protocol (str): The name of the protocol file to use for pattern matching.
                            Default is `.fls`. The file must be readable, and should contain valid match patterns.

        Raises:
            ValueError: If the root directory does not exist or is not a directory.
            OSError: If there are any issues accessing the protocol file.

        Example:
            # Create an instance of FLS with the current directory as the root
            fls = FLS(root='.', protocol='.ignorem')
        """
        self._root = os.path.normpath(os.path.abspath(root))
        if not os.path.isdir(self._root):
            raise ValueError(f"The root directory '{self._root}' is invalid or does not exist.")

        self._protocol = protocol
        self._rules = defaultdict(list)
        self._matched = {}
        self._load_rules()
        self._load_paths()

    def _load_rules(self):
        """
        Collect all relative paths and check them against the accumulated rules.

        Traverses directories from the root to the deepest level to gather
        match rules from protocol files.
        """
        for dirpath, _, filenames in sorted(os.walk(self._root), key=lambda x: x[0]):
            ancestor = dirpath if dirpath == self._root else os.path.dirname(dirpath)
            self._rules[dirpath].extend(self._rules[ancestor])

            # Process local rules if the protocol file exists
            if self._protocol in filenames:
                try:
                    self._parse_file(dirpath, os.path.join(dirpath, self._protocol))
                except Exception as e:
                    print(f"Error processing protocol file in directory '{dirpath}': {e}")

    def _parse_file(self, directory, filepath):
        """
        Load rules from the file and store them in rules.

        Args:
            directory (str): The directory containing the protocol file.
            filepath (str): The full path to the protocol file.
        """
        context = os.path.relpath(directory, self._root)

        try:
            with open(filepath, 'r') as file:
                for line in file:
                    pattern = line.strip()

                    if pattern and not pattern.startswith('#'):
                        rule = self.FLSRule(pattern, parent=self, context=context)
                        self._rules[directory].append(rule)
        except FileNotFoundError:
            print(f"File not found: The protocol file '{filepath}' does not exist.")
        except PermissionError:
            print(f"Error: Permission denied to read '{filepath}'.")
        except Exception as e:
            print(f"Error: An unexpected error occurred while reading '{filepath}': {e}")

    def _load_paths(self):
        """
        Collect all relative paths and check them against the accumulated rules.

        This method traverses the file tree and applies the rules to each file
        and directory to determine if they should be matched.
        """
        relpaths = []

        for dirpath, dirnames, filenames in sorted(os.walk(self._root), key=lambda x: x[0]):
            for dirname in dirnames:
                relpaths.append(os.path.relpath(os.path.join(dirpath, dirname), self._root) + os.sep)

            for filename in filenames:
                relpaths.append(os.path.relpath(os.path.join(dirpath, filename), self._root))

        for relpath in relpaths:
            abspath = os.path.join(self._root, relpath)
            is_matched = False

            # Check rules for the directory containing the file/directory
            directory = os.path.dirname(abspath) if abspath != self._root else self._root

            for rule in self._rules[directory]:
                if rule.match(relpath):
                    is_matched = not rule.is_negation

            # Store the path status in the dictionary
            self._matched[relpath] = is_matched

    def is_matched(self, path):
        """
        Public method for checking if a path is matched.

        Args:
            path (str): The absolute or relative path to check.

        Returns:
            bool: True if the path is matched, False otherwise.
        """
        path = os.path.normpath(os.path.abspath(path))
        path = os.path.relpath(path, self._root) + (os.sep if os.path.isdir(path) else '')

        is_matched = self._matched.get(path, None)

        return is_matched

    def matched(self):
        """
        Public method to retrieve matching results for files and directories.

        This method returns a list of tuples, where each tuple contains a relative path and a boolean value indicating
        whether the file or directory is matched based on the specified rules.

        Returns:
            dict: A dictionary where keys are relative paths, and values are booleans indicating whether each
                  path is matched (True) or ignored (False).
        """
        return self._matched.items()

    def get_all_rules(self):
        """
        Public method to retrieve all declared rules.

        This method returns a list of all rules defined for each directory.

        Returns:
            dict: A dictionary where the keys are directory paths and the values are lists of 
                  `FLSRule` objects associated with those directories.
        """
        return self._rules.items()

    def get_rules(self, directory):
        """
        Public method to retrieve rules for a specific directory.

        Args:
            directory (str): The directory to fetch rules for.

        Returns:
            list: A list of `FLSRule` objects associated with the specified directory.
        """
        return self._rules.get(directory, [])

    def add_rule(self, directory, pattern):
        """
        Optionally, add a new rule to a specific directory.

        This method adds a new match rule for the specified directory.

        Args:
            directory (str): The directory to add the rule to.
            pattern (str): The match pattern to add.
        """
        pattern = pattern.strip()
        if pattern and not pattern.startswith('#'):
            rule = self.FLSRule(pattern, parent=self)
            self._rules[directory].append(rule)


if __name__ == '__main__':
    fls = FLS(root='.', protocol='.fls')
    # Retrieve a list of matched and unmatched files and directories and print their status
    for path, is_matched in fls.matched():
        status = "Matched" if is_matched else "Ignored"
        print(f"{path}: {status}")



"""
Checking the hierarchical consistency of the rules.
"""

from env.common.run import run
from env.common.mktree import mktree


def test(tmp_path):
    mktree(path=tmp_path, dir_number=3, file_number=1, depth=3)

    if 1:  # Test 1: Checking the sequence of rule application.

        (tmp_path / 'dirA/.fls').write_text('file0.txt')

        test_cases = {
            'dirA/dirB/dirC/file0.txt': False,
            'dirA/dirB/file0.txt': False,
            'dirA/file0.txt': True,
            'dirB/file0.txt': False,
            'file0.txt': False,
            'file0.log': False,
        }

        run(tmp_path, test_cases, test_name='1')

    if 1:  # Test 2: Checking the sequence of rule application.
        (tmp_path / 'dirA/.fls').write_text('/file0.txt')

        test_cases = {
            'dirA/dirB/dirC/file0.txt': False,
            'dirA/dirB/file0.txt': False,
            'dirA/file0.txt': True,
            'dirB/file0.txt': False,
            'file0.txt': False,
            'file0.log': False,
        }

        run(tmp_path, test_cases, test_name='2')
        (tmp_path / 'dirA/.fls').write_text('')  # clear

    if 1:  # Test 3: Checking the sequence of rule application.
        (tmp_path / 'dirA/dirA/.fls').write_text('file0.txt')

        test_cases = {
            'dirA/dirA/file0.txt': True,
            'dirA/file0.txt': False,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='3')




from fls.fls import FLS
import os


# ANSI escape codes for colors
class Colors:
    BOLD = "\033[1m"
    RESET = "\033[0m"
    GREEN = "\033[32m"
    RED = "\033[31m"
    YELLOW = "\033[33m"
    CYAN = "\033[36m"
    MAGENTA = "\033[35m"


def run(test_path, test_cases, test_name):
    # Initialize the FLS object
    fls = FLS(root=test_path)

    print(f"\n{Colors.CYAN}Scan dir: {test_path}{Colors.RESET}")

    for path, is_matched in fls.matched():
        status = f"{Colors.BOLD}Matched{Colors.RESET}" if is_matched else "Ignored"
        print(f"{path}: {status}")

    for path, rules in fls.get_all_rules():
        print(f"{Colors.YELLOW}{os.path.relpath(path, test_path)}{Colors.RESET}")
        for rule in rules:
            _r = ', '.join(f"'{key}': '{Colors.CYAN}{value}{Colors.RESET}'" for key, value in rule.rule.items())
            print(f"    {_r}")

    for input, expected in test_cases.items():
        output = fls.is_matched(test_path / input)
        assert output == expected, f"{Colors.RED}Test {test_name} failed for '{input}': Expected '{expected}', got '{output}'{Colors.RESET}"



"""
Matching basic patterns.
"""

from env.common.mktree import mktree
from env.common.run import run


def test(tmp_path):

    mktree(path=tmp_path, dir_number=1, file_number=1, depth=2)

    if 1:  # Test 1: A positive match for the `file0.txt` on its context level.

        (tmp_path / '.fls').write_text('file0.txt')

        test_cases = {
            'dirA/dirA/file0.txt': False,
            'dirA/file0.txt': False,
            'file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='1')

    if 1:  # Test 2: A leading slash (/) is redundant.
        (tmp_path / '.fls').write_text('/file0.txt')

        test_cases = {
            'dirA/dirA/file0.txt': False,
            'dirA/file0.txt': False,
            'file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='2')

    if 1:  # Test 3: Matching a directory pattern.
        (tmp_path / '.fls').write_text('dirA/')

        test_cases = {
            'dirA': True,
            'dirA/dirA': False,
            'dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='3')

    if 1:  # Test 4: Checking for a trailing slashes (/) in directory matching.
        (tmp_path / '.fls').write_text('/dirA/')

        test_cases = {
            'dirA': True,
            'dirA/dirA': False,
            'dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='4')

    mktree(path=tmp_path, dir_number=1, file_number=1, depth=1)

    if 1:  # Test 5: A negative match directory pattern against a file.
        (tmp_path / '.fls').write_text('file0.txt/')

        test_cases = {
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='5')

    (tmp_path / 'dirA/dirA').write_text('Ooops! It is a file!')
    if 1:  # Test 6: A negative match file pattern against a directory.
        (tmp_path / '.fls').write_text('dirA/dirA')

        test_cases = {
            'dirA/dirA': True,
        }

        run(tmp_path, test_cases, test_name='6')

    if 1:  # Test 7:  A negative match a directory-kind pattern against a file.
        (tmp_path / '.fls').write_text('dirA/dirA/')

        test_cases = {
            'dirA/dirA': False,
        }

        run(tmp_path, test_cases, test_name='7')



"""
Brackets-based patterns test.
"""

from env.common.run import run
from env.common.mktree import mktree


def test(tmp_path):
    mktree(path=tmp_path, dir_number=3, file_number=5, depth=3)

    if 1:  # Test 1: Any file matching the pattern "file?.txt" where the "?" is a digit from 0 to 2.
        (tmp_path / '.fls').write_text('file[0-2].txt')

        test_cases = {
            'file0.txt': True,
            'file1.txt': True,
            'file2.txt': True,
        }

        run(tmp_path, test_cases, test_name='1')

    if 1:  # Test 2: Any file matching the pattern "file?.txt" where the "?" is a digit 0 or 2.
        (tmp_path / '.fls').write_text('file[02].txt')

        test_cases = {
            'file0.txt': True,
            'file1.txt': False,
            'file2.txt': True,
        }

        run(tmp_path, test_cases, test_name='2')

    if 1:  # Test 3: Any file matching the pattern "file?.txt" where the "?" is any character except those from 0 to 2.
        (tmp_path / '.fls').write_text('file[!0-2].txt')

        test_cases = {
            'file0.txt': False,
            'file1.txt': False,
            'file2.txt': False,
            'file3.txt': True,
        }

        run(tmp_path, test_cases, test_name='3')

    if 1:  # Test 4: Any file matching the pattern "file?.txt"  where the "?" is neither "0" nor "1".
        (tmp_path / '.fls').write_text('file[!01].txt')

        test_cases = {
            'file0.txt': False,
            'file1.txt': False,
            'file2.txt': True,
        }

        run(tmp_path, test_cases, test_name='4')

    if 1:  # Test 5: Any directory matching the pattern "dir?/" where the "?" is "A".
        (tmp_path / '.fls').write_text('dir[A]/')

        test_cases = {
            'dirA': True,
            'dirA/dirA': False,
            'dirB': False,
        }

        run(tmp_path, test_cases, test_name='5')

    if 1:  # Test 6: Any directory matching the pattern "dir?/", where "?" is any character except those from "C" to "Z".
        (tmp_path / '.fls').write_text('dir[!C-Z]/')

        test_cases = {
            'dirA': True,
            'dirB': True,
            'dirC': False,
        }

        run(tmp_path, test_cases, test_name='6')

    if 1:  # Test 7: Any path matching the pattern "dir?/file0.txt", where the "?" is either "A" or "C".
        (tmp_path / '.fls').write_text('dir[AC]/file0.txt')

        test_cases = {
            'dirA/file0.txt': True,
            'dirB/file0.txt': False,
            'dirC/file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='7')

    if 1:  # Test 8: Any path matching the pattern "dir?/file0.txt", where the "?" is neither "B" nor "Z".
        (tmp_path / '.fls').write_text('dir[!BZ]/file0.txt')

        test_cases = {
            'dirA/file0.txt': True,
            'dirB/file0.txt': False,
            'dirC/file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='8')



"""
Nested patterns hierarchy test.
"""

from env.common.run import run
from env.common.mktree import mktree


def test(tmp_path):
    if 1:  # Test 1: Checking the sequence of rule application.
        mktree(path=tmp_path, dir_number=2, file_number=1, depth=2)

        (tmp_path / 'dirA/.fls').write_text('file0.txt')

        test_cases = {
            'dirA/file0.txt': True,
            'dirA/dirA/file0.txt': False,
            'dirB/file0.txt': False,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='1')

    if 1:  # Test 2: Checking the sequence of rule application.
        mktree(path=tmp_path, dir_number=2, file_number=1, depth=2)

        (tmp_path / 'dirA/.fls').write_text('dirA/')

        test_cases = {
            'dirA': False,
            'dirA/dirA': True,
            'dirA/dirA/file0.txt': False,
            'dirB/dirA': False,
        }

        run(tmp_path, test_cases, test_name='2')

    if 1:  # Test 3: Checking the sequence of rule application.
        mktree(path=tmp_path, dir_number=1, file_number=1, depth=2)

        (tmp_path / '.fls').write_text('dirA/file0.txt')

        test_cases = {
            'dirA/file0.txt': True,
            'dirA/dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='3')

    if 1:  # Test 4: Checking the sequence of rule application.
        mktree(path=tmp_path, dir_number=1, file_number=0, depth=3)

        (tmp_path / '.fls').write_text('dirA/dirA/')

        test_cases = {
            'dirA/dirA': True,
            'dirA/dirA/dirA': False,
        }

        run(tmp_path, test_cases, test_name='4')

    if 1:  # Test 5: Checking the sequence of rule application.
        mktree(path=tmp_path, dir_number=1, file_number=0, depth=2)

        (tmp_path / '.fls').write_text('dirA/file0.txt')
        (tmp_path / 'dirA/.fls').write_text('!file0.txt')

        test_cases = {
            'dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='5')



"""
Asterisk-based patterns test.
"""

from env.common.run import run
from env.common.mktree import mktree


def test(tmp_path):
    mktree(path=tmp_path, dir_number=1, file_number=1, depth=2)

    if 1:  # Test 1: A positive match for any file or directory located at the root level of the context.
        (tmp_path / '.fls').write_text('*')

        test_cases = {
            'dirA': True,
            'dirA/dirA': False,
            'dirA/file0.txt': False,
            'file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='1')

    if 1:  # Test 2: The same as the previous test.
        (tmp_path / '.fls').write_text('/*')

        test_cases = {
            'dirA': True,
            'dirA/dirA': False,
            'dirA/file0.txt': False,
            'file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='2')

    if 1:  # Test 3: A positive match for any directory located at the root level of the context.
        (tmp_path / '.fls').write_text('*/')

        test_cases = {
            'dirA': True,
            'dirA/dirA': False,
            'dirA/file0.txt': False,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='3')

    if 1:  # Test 4: A trick to positively match all files at the root level of the context while excluding directories.
        (tmp_path / '.fls').write_text('*\n!*/')

        test_cases = {
            'dirA': False,
            'file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='4')

    mktree(path=tmp_path, dir_number=2, file_number=1, depth=3)

    if 1:  # Test 5: A positive match for any second-level object.
        (tmp_path / '.fls').write_text('*/*')

        test_cases = {
            'dirA': False,
            'dirA/dirA': True,
            'dirA/dirA/file0.txt': False,
            'dirA/dirB': True,
            'dirA/file0.txt': True,
            'dirB': False,
            'dirB/dirA': True,
            'dirB/dirB': True,
        }

        run(tmp_path, test_cases, test_name='5')

    if 1:  # Test 6: A positive match for any second-level directory.
        (tmp_path / '.fls').write_text('*/*/')

        test_cases = {
            'dirA': False,
            'dirB': False,
            'dirA/dirA': True,
            'dirA/dirA/file0.txt': False,
            'dirA/file0.txt': False,
            'dirA/dirB': True,
            'dirB/dirA': True,
            'dirB/dirB': True,
            'dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='6')

    if 1:  # Test 7: A pattern to match any object located at the third-level within the context.
        (tmp_path / '.fls').write_text('*/*/*')

        test_cases = {
            'dirA/dirA': False,
            'dirA/dirA/dirA': True,
            'dirA/dirA/dirA/file0.txt': False,
            'dirA/dirA/file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='7')

    if 1:  # Test 8: A pattern to match any third-level directory.

        (tmp_path / '.fls').write_text('*/*/*/')

        test_cases = {
            'dirA/dirA': False,
            'dirA/dirA/dirA': True,
            'dirA/dirA/dirA/file0.txt': False,
            'dirA/dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='8')

    if 1:  # Test 9: Match files located directly inside any first-level subdirectory of "dirA".
        mktree(path=tmp_path, dir_number=2, file_number=1, depth=3)

        (tmp_path / '.fls').write_text('dirA/*/file0.txt')

        test_cases = {
            'dirA/file0.txt': False,
            'dirA/dirA/file0.txt': True,
            'dirA/dirB/file0.txt': True,
            'dirA/dirA/dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='9')

    if 1:  # Test 10: A positive match for any file or directory ending with "txt")
        (tmp_path / '.fls').write_text('*txt')

        test_cases = {
            'file0.txt': True,
            'dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='10')

    if 1:  # Test 11:  A positive match for any file or directory starting with "file".
        (tmp_path / '.fls').write_text('file*')

        test_cases = {
            'dirA/file0.txt': False,
            'dirA/file0.log': False,
            'file0.txt': True,
            'file0.log': True,
        }

        run(tmp_path, test_cases, test_name='11')

    if 1:  # Test 12:  A positive match for any directory starting with "dir".
        (tmp_path / '.fls').write_text('dir*/')

        test_cases = {
            'dirA': True,
            'dirA/dirA': False,
            'dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='12')

    if 1:  # Test 13:  It matches any object located directly inside the `dirA` directory, excluding the `dirA` itself.
        (tmp_path / '.fls').write_text('dirA/*')

        test_cases = {
            'dirA': False,
            'dirA/dirA': True,
            'dirA/file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='13')

    if 1:  # Test 14:  It matches any dot-files.
        (tmp_path / '.fls').write_text('.*')

        test_cases = {
            '.fls': True,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='14')



"""
Exclamation-mark-based patterns test.
"""

from env.common.run import run
from env.common.mktree import mktree


def test(tmp_path):
    mktree(path=tmp_path, dir_number=3, file_number=2, depth=3)

    if 1:  # Test 1.
        (tmp_path / '.fls').write_text('file0.txt')
        (tmp_path / 'dirA/.fls').write_text('!file0.txt')

        test_cases = {
            'dirA/dirA/file0.txt': False,
            'dirA/file0.txt': False,
            'file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='1')

        (tmp_path / 'dirA/.fls').write_text('')

    if 1:  # Test 2.
        (tmp_path / '.fls').write_text('!file0.txt\n!file1.log')
        (tmp_path / 'dirA/dirA/.fls').write_text('file0.txt')
        (tmp_path / 'dirB/dirB/.fls').write_text('file1.log')

    if 1:  # Test 3.
        (tmp_path / '.fls').write_text('*\n!file*.*')

        test_cases = {
            'dirA': True,
            'dirB': True,
            'file0.txt': False,
            'file0.log': False,
        }

        run(tmp_path, test_cases, test_name='3')

        (tmp_path / 'dirA/dirB/.fls').write_text('')

    if 1:  # Test 4.
        (tmp_path / '.fls').write_text('**')
        (tmp_path / 'dirA/dirA/.fls').write_text('!**')

        test_cases = {
            'dirA': True,
            'dirA/dirA': True,
            'dirA/dirA/dirA': False,
            'dirA/dirA/dirA/file0.txt': False,
            'dirA/dirA/file0.txt': False,
            'dirA/file0.txt': True,
            'file0.log': True,
        }

        run(tmp_path, test_cases, test_name='4')



"""
Replace three or more consecutive non-escaped asterisks (*) with two.
"""

import re
import string
import random


def clean_pattern(pattern):
    placeholder_map = {}

    def _hidescape(match):
        placeholder = ''.join(random.choices(string.ascii_letters + string.digits, k=6))
        placeholder_map[placeholder] = match.group(0)

        return placeholder

    pattern = re.sub(r'\\(.)', _hidescape, pattern)
    pattern = re.sub(r'\*{3,}', '**', re.sub(r'\\(.)', r'\\\1', pattern))

    for placeholder, value in placeholder_map.items():
        pattern = re.sub(re.escape(placeholder), value, pattern)

    return pattern


test_cases = {
    '***foobar': '**foobar',
    'foo***bar': 'foo**bar',
    'foobar***': 'foobar**',
    '****foo****bar****': '**foo**bar**',
    '***/foo/***/bar/***': '**/foo/**/bar/**',
    'foo\***': 'foo\***',
}

for input, expected in test_cases.items():
    output = clean_pattern(input)
    assert output == expected, f"Test failed for '{input}': Expected '{expected}', got '{output}'"

print("All tests passed!")



"""
Question-mark-based patterns test.
"""

from env.common.run import run
from env.common.mktree import mktree


def test(tmp_path):
    mktree(path=tmp_path, dir_number=2, file_number=2, depth=3)

    if 1:  # Test 1: A positive match for filename where the "?" represents exactly one character excepting a slash (/).
        (tmp_path / '.fls').write_text('file?.txt')

        test_cases = {
            'file0.txt': True,
            'file1.txt': True,
        }

        run(tmp_path, test_cases, test_name='1')

    if 1:  # Test 2:  A negative match because "the `?` represents exactly one character".
        (tmp_path / '.fls').write_text('file0?.txt')

        test_cases = {
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='2')

    if 1:  # Test 3:  A negative match because "other than a slash (/)"..
        (tmp_path / '.fls').write_text('dirA?file0.txt')

        test_cases = {
            'dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='3')

    if 1:  # Test 4: A positive match of multiple question marks.
        (tmp_path / '.fls').write_text('f???????t')

        test_cases = {
            'file0.txt': True,
            'file1.txt': True,
        }

        run(tmp_path, test_cases, test_name='4')



"""
Escaped patterns test.
"""

import os

from env.common.run import run
from env.common.mktree import mktree


def test(tmp_path):
    # Create test directories and .fls rules
    mktree(path=tmp_path, dir_number=1, file_number=1, depth=1)

    if 1:  # Test 1: A positive match for an escaped asterisk (*).
        (tmp_path / '.fls').write_text(r'\*')
        (tmp_path / '*').write_text('')

        test_cases = {
            '*': True,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='1')

    if 1:  # Test 2: A positive match for an escaped double asterisk (**).
        (tmp_path / '.fls').write_text(r'\*\*')
        (tmp_path / '**').write_text('')

        test_cases = {
            '**': True,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='2')

    if 1:  # Test 3: Another positive match for an escaped double asterisk (**).
        (tmp_path / '.fls').write_text(r'file\*\*.txt')
        (tmp_path / 'file**.txt').write_text('')

        test_cases = {
            'file**.txt': True,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='3')

    if 1:  # Test 4: One more positive match for an escaped asterisk used as a directory contents (*).
        (tmp_path / '.fls').write_text(r'dirA/\*/')
        os.makedirs(tmp_path / 'dirA/*', exist_ok=True)

        test_cases = {
            'dirA': False,
            'dirA/*': True,
        }

        run(tmp_path, test_cases, test_name='4')

    if 1:  # Test 5: A positive match for an escaped exclamation mark (!).
        (tmp_path / '.fls').write_text(r'\!file.txt')
        (tmp_path / '!file.txt').write_text('')

        test_cases = {
            '!file.txt': True,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='5')

    if 1:  # Test 6: A positive match for an escaped hash (#)
        (tmp_path / '.fls').write_text(r'\#file.txt')
        (tmp_path / '#file.txt').write_text('')

        test_cases = {
            '#file.txt': True,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='6')



"""
Double-asterisk-based patterns test.
"""

import os

from env.common.run import run
from env.common.mktree import mktree


def test(tmp_path):
    mktree(path=tmp_path, dir_number=2, file_number=2, depth=4)

    if 1:  # Test 1: A positive match for any object recursively.
        (tmp_path / '.fls').write_text('**')

        test_cases = {
            'dirA': True,
            'dirA/dirA': True,
            'dirA/dirA/dirA': True,
            'dirA/dirA/file0.txt': True,
            'dirA/dirA/dirA/file0.txt': True,
            'file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='1')

    if 1:  # Test 2: A positive match for any directoriy recursively.
        (tmp_path / '.fls').write_text('**/')

        test_cases = {
            'dirA': True,
            'dirA/dirA': True,
            'dirA/dirA/dirA': True,
            'dirA/dirA/dirA/file0.txt': False,
            'dirA/dirA/file0.txt': False,
            'dirA/file0.txt': False,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='2')

    if 1:  # Test 3: Positively match all files at the root level of the context while excluding directories.
        (tmp_path / '.fls').write_text('**\n!**/')

        test_cases = {
            'dirA': False,
            'dirA/dirA': False,
            'dirA/dirA/dirA': False,
            'dirA/dirA/dirA/file0.txt': True,
            'dirA/dirA/file0.txt': True,
            'dirA/file0.txt': True,
            'file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='3')

    if 1:  # Test 4: Match any object inside the `dirA` directory, recursively, excluding the `dirA` directory itself.
        (tmp_path / '.fls').write_text('dirA/**')

        test_cases = {
            'dirA': False,
            'dirA/dirA': True,
            'dirA/dirA/dirA': True,
            'dirA/dirA/dirA/file0.txt': True,
            'dirA/dirA/file0.txt': True,
            'dirA/file0.txt': True,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='4')

    if 1:  # Test 5: A pattern to recursively match any path starting with "dir" and ending with "0.txt"
        (tmp_path / '.fls').write_text('dir**txt')
        (tmp_path / 'dirAfoo0.txt').write_text('a file')
        os.makedirs(tmp_path / 'dirA/bar0.txt', exist_ok=True)

        test_cases = {
            'dirA/dirA/file0.txt': True,
            'dirA/dirA/file1.txt': True,
            'dirA/dirB/file0.txt': True,
            'dirA/dirB/file1.txt': True,
            'dirA/file0.txt': True,
            'dirB/dirA/file0.txt': True,
            'dirAfoo0.txt': True,  # a file
            'dirA/bar0.txt': True,  # a directory
        }

        run(tmp_path, test_cases, test_name='5')

    if 1:  # Test 6: It will not match "dirA/file0.txt" because "/**/" requires at least one additional level of
        #            nesting between dirA and "file0.txt".
        (tmp_path / '.fls').write_text('dirA/**/file0.txt')

        test_cases = {
            'dirA/dirA/file0.txt': True,
            'dirA/dirA/dirA/file0.txt': True,
            'dirA/dirB/file0.txt': True,
            'dirA/file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='6')

    if 1:  # Test 7: In contrast, the pattern matches dirA/file0.txt here, as /** allows matching files at any depth
        #            within dirA, including directly inside it. The slashes make the difference!
        (tmp_path / '.fls').write_text('dirA/**file0.txt')

        test_cases = {
            'dirA/dirA/file0.txt': True,
            'dirA/dirA/dirA/file0.txt': True,
            'dirA/dirB/file0.txt': True,
            'dirA/file0.txt': True,
        }

        run(tmp_path, test_cases, test_name='7')

    if 1:  # Test 8: A recursive match for all objects located at the second level and deeper.
        (tmp_path / '.fls').write_text('**/**')

        test_cases = {
            'dirA': False,
            'dirA/dirA': True,
            'dirA/dirA/dirA': True,
            'dirA/dirA/dirA/dirA': True,
            'dirA/dirA/dirA/dirA/file0.txt': True,
            'dirA/dirA/dirA/file0.txt': True,
            'dirA/dirA/file0.txt': True,
            'dirA/file0.txt': True,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='8')

    if 1:  # Test 9: A recursive match for all directories located at the second level and deeper.
        (tmp_path / '.fls').write_text('**/**/')

        test_cases = {
            'dirA': False,
            'dirA/dirA': True,
            'dirA/dirA/dirA': True,
            'dirA/dirA/dirA/dirA': True,
            'dirA/dirA/dirA/dirA/file0.txt': False,
            'dirA/dirA/dirA/file0.txt': False,
            'dirA/dirA/file0.txt': False,
            'dirA/file0.txt': False,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='9')

    if 1:  # Test 10: A recursive match for all objects located at the second level and deeper.
        (tmp_path / '.fls').write_text('**/**/**')

        test_cases = {
            'dirA': False,
            'dirA/dirA': False,
            'dirA/dirA/dirA': True,
            'dirA/dirA/dirA/dirA': True,
            'dirA/dirA/dirA/dirA/file0.txt': True,
            'dirA/dirA/dirA/file0.txt': True,
            'dirA/dirA/file0.txt': True,
            'dirA/file0.txt': False,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='10')

    if 1:  # Test 11: A recursive match for all directories located at the third level and deeper.
        (tmp_path / '.fls').write_text('**/**/**/')

        test_cases = {
            'dirA': False,
            'dirA/dirA': False,
            'dirA/dirA/dirA': True,
            'dirA/dirA/dirA/dirA': True,
            'dirA/dirA/dirA/dirA/file0.txt': False,
            'dirA/dirA/dirA/file0.txt': False,
            'dirA/dirA/file0.txt': False,
            'dirA/file0.txt': False,
            'file0.txt': False,
        }

        run(tmp_path, test_cases, test_name='11')



from setuptools import setup, find_packages

setup(
    name="FLS",
    version="0.0.1.241224",
    description="A module for matching file system paths against patterns.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Andrii Burkatskyi aka andr11b",
    author_email="4ndr116@gmail.com",
    url="https://github.com/codyverse/fls",
    license="MIT",
    packages=find_packages(include=["fls"]),
    install_requires=[],
    extras_require={
        "dev": [
            "pytest>=6.0",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.8",
)



"""
Remove redundant slashes (/).
"""
import re


def clean_pattern(pattern):
    return (re.sub(r'/+', '/', pattern)).lstrip('/')


test_cases = {
    '/foo/bar': 'foo/bar',
    'foo/bar': 'foo/bar',
    'foo/bar/': 'foo/bar/',
    '//foo/bar/': 'foo/bar/',
    'foo//bar/': 'foo/bar/',
    'foo/bar//': 'foo/bar/',
    '///foo/bar': 'foo/bar',
    'foo///bar': 'foo/bar',
    'foo/bar///': 'foo/bar/',
    '///foo///bar///': 'foo/bar/',
}

for input, expected in test_cases.items():
    output = clean_pattern(input)
    assert output == expected, f"Test failed for '{input}': Expected '{expected}', got '{output}'"

print("All tests passed!")



import os
import string
import shutil

def mktree(path, dir_number=1, file_number=1, depth=1):
    """
    Creating a specified 'depth' of nesting levels with 'dir_number' directories (dirA to dirZ)
    and 'file_number' files (file0.log to file9.log, file0.txt to file9.txt).
    """
    def create_files(path, files):
        for file_name in files:
            file_path = os.path.join(path, file_name)
            with open(file_path, 'w') as f:
                f.write(f'This is {file_name}\n')  # Optional content

    # Safely remove the base directory if it exists
    if os.path.exists(path):
        shutil.rmtree(path)

    # Limit the number of directories to 26 (A-Z) and files to 10 (0-9)
    dir_number = (dir_number - 1) % 26 + 1
    file_number = (file_number - 1) % 10 + 1

    # Generate directory names (dirA, dirB, ..., dirZ)
    dir_names = [f'dir{letter}' for letter in string.ascii_uppercase[:dir_number]]

    # Generate file names (file0.txt, file1.txt, ..., file9.txt and file0.log, file1.log, ..., file9.log)
    txt_files = [f'file{digit}.txt' for digit in range(file_number)]
    log_files = [f'file{digit}.log' for digit in range(file_number)]

    # Create the directories and files in the base path
    for dir_name in dir_names:
        current_dir_path = os.path.join(path, dir_name)
        os.makedirs(current_dir_path, exist_ok=True)

        # Create .txt and .log files in the current directory
        create_files(current_dir_path, txt_files)
        create_files(current_dir_path, log_files)

        # If depth is greater than 1, create nested directories and files
        if depth > 1:
            mktree(current_dir_path, dir_number, file_number, depth - 1)

    # Create .txt and .log files in the base path as well
    create_files(path, txt_files)
    create_files(path, log_files)

# Example usage
#mktree('test_structure', dir_number=5, file_number=3, depth=2)



import os
import glob
import shutil
import importlib.util
import pytest
import sys
from pathlib import Path

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))


def clean_pycache(directory):
    """Remove __pycache__ directories."""

    for dirpath, dirnames, _ in os.walk(directory):
        if '__pycache__' in dirnames:
            pycache_path = os.path.join(dirpath, '__pycache__')
            shutil.rmtree(pycache_path)


def run_tests():
    """Run all tests in the `tests` directory."""

    environment = 'env'
    test_modules = []  # List to keep track of test modules

    # Clean up `__pycache__` before running tests
    clean_pycache(environment)

    for test_file in sorted(glob.glob(os.path.join(environment, 'tests/fn', 'test_*.py'))):
        if os.path.exists(test_file):
            spec = importlib.util.spec_from_file_location("test", test_file)
            test_module = importlib.util.module_from_spec(spec)

            # Add the loaded test module to the list
            test_modules.append(test_module)

    # After running all setup scripts, run pytest to execute all tests
    if test_modules:  # Only run pytest if there are test modules
        pytest.main(['-s', '--tb=line', *[test.__file__ for test in test_modules]])

    # Clean up __pycache__ after running tests
    clean_pycache(environment)


if __name__ == '__main__':
    run_tests()
